#!/usr/local/bin/perl
##############################################################################
# Author: Binny V Abraham                                                    #
# Website: http://www.geocities.com/binnyva                                  #
# E-Mail : binnyva@hotmail.com                                               #
# Get more Perl scripts from http://www.geocities.com/binnyva/code/perl      #
##############################################################################
# Name    : Source Compressor                                                #
# Version : 1.00.A                                                           #
# Date    : August 12, 2005                                                  #
# Compresses the Javascript source to a file to a much smaller file. Replaces#
# 	large variable names with smaller,generic names and removes unwanted     #
#   white spaces.                                                            #
#                                                                            #
# Get more Perl scripts from http://www.geocities.com/binnyva/code/perl      #
##############################################################################

#Includes
# use strict;
# use warnings;
use Getopt::Long;

############################## Global Variables ##########################################
#Program Information
my %program = (
	name	=>	"Source Compressor",
	author	=>	"Binny V Abraham",
	version	=>	"1.00.A",
	email	=>	"binnyva\@hotmail.com",
	website	=>	"http://www.geocites.com/binnyva"
);

my @vars;
my %var_exists = (); #Name of all the variables
my $chars = "abcdefghijklmnopqrstuvwxyz";
my $var_name = "";
my $file_name = "";

#Settings - Set this using the Command line arguments - like 
#	perl "compresser.pl" -display -verbose
my $display = 1;
my $verbose = 0;
my $readable = 0; #Use \n instead of ';' as the command seperator.
my $change_name = 0; #Rename the variables to a shorter version
my $remove_vars = 0; #Remove the 'var' keyword - this is asking for trouble
my $print_to_screen = 0; #Print the compressed data - instead of writing it to a file.
my $safe_replace = 0;

# #Get the command line options
GetOptions ('display:0'		=> \$display,
            'verbose:1'		=> \$verbose,
            'readable:1'	=> \$readable,
            'change-name:1'	=> \$change_name,
            'remove-vars:1'	=> \$remove_vars,
            'print:1'		=> \$print_to_screen,
            'safe:1'		=> \$safe_replace
);

$file_name = $ARGV[0] or usage();

$display = 0 if $print_to_screen; # Disable the program message if the compressed data is printed

if ($display) {
	print "$program{'name'} $program{'version'}\n";
	print "Compressing $file_name...\n";
}
#Open the file and get all the contents
open(IN,$file_name) or die "Cant open file $file_name : $!";
my @all_lines = <IN>;
my $contents = join "",@all_lines;

#Get all the variable names and concatinate all the data to a single variable.
foreach (@all_lines) {
	if(/var (\w+)/) {#Search for all variable names
		#Find wether the variable name is in the array.
		unless ($var_exists{$1}) {
    		$var_exists{$1} = 1;
    		push(@vars,$1);
		}
	}
}

if($change_name and !$safe_replace) {
	#Replace all variable names with shorter versions
	my $new_name = "";
	foreach my $name (@vars) {
	 	my $new_name = getNewVariableName();
	 	$contents =~ s/(\b)$name(\b)/$1$new_name$2/g;
		print "Changed '$name' -> '$new_name'.\n" if($verbose and $display);
	}
	print "$#vars variable names changed.\n" if $display;
}

############################## RegExp Replaces ######################################

#Remove //... type of comments
$contents =~ s/\s*\/\/.*?\n/\n/g;
#Replace /* ... */ type of comments - dont put this at the top - won't work for some reason.
$contents =~ s!/\*.*?\*/\s*!!sg;

$contents =~ s/(;|\n)\t+/$1/g;#The whitespaces at the start of the line
$contents =~ s/\n{2,}/\n/g;#Remove multiple \n's with a single one

#Remove some commely used 'clean code' white spaces
$contents =~ s/(if|else)\s*(\(.*?\))?\s*([\w\{])/$1$2 $3/sg; #If the statements are in the next line.
#The following code will be replaced
#'a = b + c;' will be changed to 'a=b+c;'
#'if ( a >= 0 ) { ' -> 'if(a>=0){'
#'if ( a == 0 || b >= 0 && c <= 0 ){' -> 'if(a==0||b>=0&&c<=0){'
# and more...
$contents =~ s/[ \t]*([\=\,\+\-\*\&\|\:\?\<\>\{\}\(\)\[\]]+)[ \t]*/$1/g;


unless($safe_replace) {
	print "Doing Unsafe compression.\n" if ($display);
	
	#Put in the necessary ';' before replacing the \n's
	$contents =~ s/([\w\"\'\)])\s*\n/$1;/g;
	$contents =~ s/\n//g;

	if($readable) {
		#Use \n's whereever possible - improves readabilty of the compressed script. - Why?
		$contents =~ s/;/\n/g;
		$contents =~ s/for\((.*?)\n(.*?)\n(.*?)\)/for($1;$2;$3)/g;#The for loops need their ';' chars.
	}

	if($remove_vars) {
		$contents =~ s/(\b)var /$1/g;# Remove non critical words - like 'var'
	}
}

print "File Compressed.\n" if $display;

if($print_to_screen) {
	print $contents;
} else {
	#Get the file title and extension.
	my $dot_pos = rindex($file_name,".");
	my $file_title = substr($file_name,0,$dot_pos);
	my $file_extension = substr($file_name,$dot_pos+1);
	my $new_file_name = $file_title . "_compressed." . $file_extension;
	
	#Write the compressed data to the new file.
	open(OUT,">$new_file_name") or die "Can't open $new_file_name for writing : $!";
	print OUT $contents;
	close OUT;
	print "Wrote the compressed data to $new_file_name.\n" if $display;
}

############################## Functions ##########################################

#Get the next variable name - if the current variable is 'a', returns 'b'. 
#		If the current one is 'z', returns 'aa' and so on. The actual calculation of the variable is done
#		in the nextName() function - but that is a recurisve funtion - so this function will act as a 
#		front end for that function.
sub getNewVariableName {
	if($var_name eq "") {#This happens only once - for the very first variable.
		$var_name = "a";
	}
	else {
		$var_name = nextName($var_name);
	}
	return $var_name;
}

#The recursive function - see the comments on getNewVariableName() function for more info.
sub nextName {
	my ($name) = @_; #Argument
	my $last_char = substr($name,-1);#Get the last char
	my ($index,$new_last_char);

	if($last_char eq "z") {#Once we have reached z, start from a again
		$name = nextName(substr($name,0,-1)) if(length($name) > 1);#This function will continue recrusiveley - until all 'z' are cleared
		$new_last_char = substr($name,-1);#Get the last char
		$new_last_char = "a" if($new_last_char eq "z"); #If the last char is 'z', the new last char is 'a';
	}
	else {
		#Find the increment of the last char
		$index = index($chars,$last_char);
		$new_last_char = substr($chars,$index+1,1);
	}

	substr($name,-1,1,$new_last_char);#Replace the last charector with the '$new_last_char'
	
	$name .= "a" if($last_char eq "z"); #If we have reached z, add new char to the end.
	
	return $name;
}

#Display a help message
sub usage {
	print <<END;
$program{'name'} $program{'version'}
By $program{'author'} - $program{'website'}

$program{'name'} is a perl script that can be used to compress the Javascript 
source file to a much smaller file. Replaces large variable names with smaller,
generic names and removes unwanted white spaces.

Usage
-----
perl compressor.pl <JS File> [options]
Example :
perl compressor.pl settings.js -display -verbose -change-name

Command Line Options
--------------------
The first argument MUST be the file name of the JavaScript file.
-verbose		Show details of what is going on.
-readable		Use \\n instead of ';' as the command seperator.
-change-name		Rename the variables to a shorter version. Enabled by default.
-remove-vars		Remove the 'var' keyword - this is asking for trouble.
-print			Outputs the compressed data instead of writing it to a file.
END

exit -1;
}

############################## To Do ##########################################
# Changes the text if the variable name appears in the coments/strings/other places.
# The variables appearing more times must have shorter number of chars.
# Change the variables that don't use the 'var' decleration method. - Is this possible?
# Implement Unremoveable comments - needed for copyright statements.
# Change the names of the functions/classes too?
